from game.models import Attempt


def create_attempt(student, level, score):
    attempt = Attempt(level=level, student=student, score=score, is_best_attempt=True)
    attempt.save()

    return attempt



from django.db import migrations


def update_episodes(apps, schema_editor):
    Level = apps.get_model("game", "Level")
    episode7_levels = list(Level.objects.filter(name__range=(51, 60)))
    del episode7_levels[0]
    episode8_levels = list(Level.objects.filter(name__range=(61, 67)))

    Episode = apps.get_model("game", "Episode")
    episode6 = Episode.objects.get(id=6)
    episode7 = Episode.objects.get(id=7)
    episode8 = Episode.objects.get(id=8)
    episode9 = Episode.objects.get(id=9)

    episode7.name = "Limited Blocks"
    episode8.name = "Procedures"

    episode7.save()
    episode8.save()

    episode6.next_episode = episode7
    episode7.next_episode = episode8
    episode8.next_episode = episode9

    episode6.save()
    episode7.save()
    episode8.save()

    episode7.level_set.set(episode7_levels)
    episode8.level_set.set(episode8_levels)

    episode7.save()
    episode8.save()


class Migration(migrations.Migration):
    dependencies = [("game", "0067_level_score_27")]

    operations = [
        migrations.RunPython(update_episodes, reverse_code=migrations.RunPython.noop)
    ]



import csv

from django.http import HttpResponse

Headers = [
    "Class",
    "Name",
    "Completed Levels",
    "Total Time",
    "Total Score",
]


def scoreboard_csv(student_data, requested_sorted_levels, improvement_data):
    response = HttpResponse(content_type="text/csv")
    response["Content-Disposition"] = 'attachment; filename="scoreboard.csv"'

    improvement_dict = {}
    for student in improvement_data:
        idx = f"{student.class_field.name}_{student.name}"
        improvement_dict[idx] = student.areas

    headers = Headers + requested_sorted_levels + ["Areas for improvement"]

    rows = []
    for student in student_data:
        result = [
            student.class_field.name,
            student.name,
            student.completed,
            student.total_time,
            student.total_score,
        ]

        scores = []
        for level_id, level_score in student.level_scores.items():
            scores.append(level_score["score"])

        row = result + scores

        # add the areas to improve, if there's any
        idx = f"{student.class_field.name}_{student.name}"
        row.append(improvement_dict.get(idx))

        rows.append(row)

    writer = csv.writer(response)
    writer.writerow(headers)
    writer.writerows(rows)

    return response



from game.end_to_end_tests.base_game_test import BaseGameTest


class TestCrashes(BaseGameTest):
    def test_crash_turning_left_on_straight_road(self):
        self.run_crashing_test(
            level=6, workspace_file="crash_turning_left_on_straight_road"
        )

    def test_crash_turning_left_on_right_turn(self):
        self.run_crashing_test(
            level=6, workspace_file="crash_turning_left_on_right_turn"
        )

    def test_crash_turning_right_on_straight_road(self):
        self.run_crashing_test(
            level=6, workspace_file="crash_turning_right_on_straight_road"
        )

    def test_crash_turning_right_on_left_turn(self):
        self.run_crashing_test(
            level=6, workspace_file="crash_turning_right_on_left_turn"
        )

    def test_crash_going_forward_on_right_turn(self):
        self.run_crashing_test(
            level=6, workspace_file="crash_going_forward_on_right_turn"
        )

    def test_crash_going_forward_on_left_turn(self):
        self.run_crashing_test(
            level=6, workspace_file="crash_going_forward_on_left_turn"
        )

    def test_crash_going_forward_on_t_junction(self):
        self.run_crashing_test(
            level=13, workspace_file="crash_going_forward_on_t_junction"
        )

    def test_crash_going_left_on_t_junction(self):
        self.run_crashing_test(
            level=40, workspace_file="crash_going_left_on_t_junction"
        )

    def test_crash_going_right_on_t_junction(self):
        self.run_crashing_test(
            level=40, workspace_file="crash_going_right_on_t_junction"
        )

    def test_running_out_of_instructions(self):
        self.running_out_of_instructions_test(
            level=6, workspace_file="running_out_of_instructions"
        )

    def test_running_out_of_fuel(self):
        self.running_out_of_fuel_test(level=68, workspace_file="running_out_of_fuel")

    def test_running_a_red_light(self):
        self.running_a_red_light_test(level=44, workspace_file="running_a_red_light")

    def test_not_delivered_everywhere(self):
        self.not_delivered_everywhere_test(
            level=16, workspace_file="not_delivered_everywhere"
        )

    def test_procedure_undefined(self):
        self.undefined_procedure_test(level=63, workspace_file="undefined_procedure")

    def test_crash_turning_left_clear_and_pass_level(self):
        page = (
            self.run_crashing_test(
                level=6, workspace_file="crash_turning_left_on_straight_road"
            )
            .try_again()
            .clear()
        )

        self.complete_and_check_level(6, page)



from django.db import migrations


def release_levels(apps, schema_editor):
    Episode = apps.get_model("game", "Episode")
    for episode in Episode.objects.all():
        episode.in_development = False
        episode.save()


class Migration(migrations.Migration):
    dependencies = [("game", "0038_level_score_40")]

    operations = [migrations.RunPython(release_levels)]



from __future__ import division
from __future__ import absolute_import

from builtins import str
from django.core.cache import cache
from django.db.models import Max
from django.shortcuts import render
from django.utils.safestring import mark_safe

import game.level_management as level_management
import game.messages as messages
from game import app_settings
from game import random_road
from game.cache import cached_episode
from game.models import Attempt, Episode
from .level_editor import play_anonymous_level


def max_score(level):
    score = 0
    if not level.disable_route_score:
        score = score + 10
    if level.model_solution != "[]":
        score = score + 10
    return score


def fetch_episode_data_from_database(early_access, start, end):
    episode_data = []
    episode = Episode.objects.get(pk=start)
    current = start
    while episode is not None:
        if episode.in_development and not early_access:
            break

        levels = []
        minName = None
        maxName = None
        for level in episode.levels:
            level_name = int(level.name)
            if not maxName or level_name > maxName:
                maxName = level_name
            if not minName or level_name < minName:
                minName = level_name

            levels.append(
                {
                    "id": level.id,
                    "name": level_name,
                    "maxScore": max_score(level),
                    "title": get_level_title(level_name),
                }
            )

        e = {
            "id": episode.id,
            "levels": levels,
            "first_level": minName,
            "last_level": maxName,
            "random_levels_enabled": episode.r_random_levels_enabled,
        }

        episode_data.append(e)
        episode = episode.next_episode

        current += 1
        if current > end:
            break

    return episode_data


def fetch_episode_data(early_access, start=1, end=12):
    key = f"episode_data_{start}_{end}"
    if early_access:
        key = f"episode_data_early_access_{start}_{end}"

    data = cache.get(key)
    if data is None:
        data = fetch_episode_data_from_database(early_access, start, end)
        cache.set(key, data)
    return [
        dict(
            episode,
            name=messages.get_episode_title(episode["id"]),
            levels=[
                dict(level, title=get_level_title(level["name"]))
                for level in episode["levels"]
            ],
        )
        for episode in data
    ]


def get_level_title(i):
    title = "title_level" + str(i)
    try:
        titleCall = getattr(messages, title)
        return mark_safe(titleCall())
    except AttributeError:
        return ""


def attach_attempts_to_level(attempts, level):
    level["score"] = attempts.get(level["id"])


def is_student(user):
    return hasattr(user.userprofile, "student")


def levels(request):
    """Loads a page with all levels listed.

    **Context**

    ``RequestContext``
    ``episodes``

    **Template:**

    :template:`game/level_selection.html`
    """
    user = request.user
    if user.is_authenticated and is_student(user):
        best_attempts = (
            Attempt.objects.filter(student=user.userprofile.student)
            .values("level_id")
            .annotate(best_score=Max("score"))
            .all()
        )

        attempts = {a["level_id"]: a["best_score"] for a in best_attempts}
    else:
        attempts = {}

    blockly_episodes = fetch_episode_data(
        app_settings.EARLY_ACCESS_FUNCTION(request), 1, 9
    )
    for episode in blockly_episodes:
        for level in episode["levels"]:
            attach_attempts_to_level(attempts, level)

    python_episodes = fetch_episode_data(
        app_settings.EARLY_ACCESS_FUNCTION(request), 10, 11
    )
    for episode in python_episodes:
        for level in episode["levels"]:
            attach_attempts_to_level(attempts, level)

    owned_level_data = []
    shared_level_data = []
    if not request.user.is_anonymous:
        owned_levels, shared_levels = level_management.get_loadable_levels(request.user)

        for level in owned_levels:
            owned_level_data.append(
                {
                    "id": level.id,
                    "title": level.name,
                    "score": attempts.get(level.id),
                    "maxScore": 10,
                }
            )

        for level in shared_levels:
            shared_level_data.append(
                {
                    "id": level.id,
                    "title": level.name,
                    "owner": level.owner.user,
                    "score": attempts.get(level.id),
                    "maxScore": 10,
                }
            )

    return render(
        request,
        "game/level_selection.html",
        context={
            "blocklyEpisodes": blockly_episodes,
            "pythonEpisodes": python_episodes,
            "owned_levels": owned_level_data,
            "shared_levels": shared_level_data,
            "scores": attempts,
        },
    )


def random_level_for_episode(request, episodeID):
    """Generates a new random level based on the episodeID

    Redirects to :view:`game.views.play_level` with the id of the newly created :model:`game.Level`.
    """
    episode = cached_episode(episodeID)
    level = random_road.create(episode)
    return play_anonymous_level(request, level.id, False, random_level=True)



from __future__ import unicode_literals

from django.db import migrations
from game.level_management import set_decor_inner, set_blocks_inner
import json


def change_levels(apps, schema_editor):

    Level = apps.get_model("game", "Level")
    LevelDecor = apps.get_model("game", "LevelDecor")
    LevelBlock = apps.get_model("game", "LevelBlock")
    Block = apps.get_model("game", "Block")

    def set_decor(level, decor):
        set_decor_inner(level, decor, LevelDecor)

    def set_blocks(level, blocks):
        set_blocks_inner(level, blocks, LevelBlock, Block)

    level61 = Level.objects.get(name="61", default=1)
    level61.model_solution = "[9]"
    level61.save()

    level75 = Level.objects.get(name="75", default=1)
    level73 = Level.objects.get(name="73", default=1)
    level72 = Level.objects.get(name="72", default=1)
    level74 = Level.objects.get(name="74", default=1)
    level79 = Level.objects.get(name="79", default=1)

    level72.next_level = level75
    level75.next_level = level74
    level74.next_level = level73
    level73.next_level = level79
    level79.next_level = None

    level74.path = (
        '[{"coordinate":[9,2],"connectedNodes":[1]},'
        + '{"coordinate":[8,2],"connectedNodes":[2,9,0]},'
        + '{"coordinate":[7,2],"connectedNodes":[1,3]},'
        + '{"coordinate":[7,1],"connectedNodes":[4,2]},'
        + '{"coordinate":[6,1],"connectedNodes":[5,3,6]},'
        + '{"coordinate":[5,1],"connectedNodes":[8,4,34]},'
        + '{"coordinate":[6,0],"connectedNodes":[4]},'
        + '{"coordinate":[3,2],"connectedNodes":[28,37,35]},'
        + '{"coordinate":[4,1],"connectedNodes":[35,5]},'
        + '{"coordinate":[8,3],"connectedNodes":[10,1]},'
        + '{"coordinate":[8,4],"connectedNodes":[11,9]},'
        + '{"coordinate":[8,5],"connectedNodes":[12,33,10]},'
        + '{"coordinate":[8,6],"connectedNodes":[13,11]},'
        + '{"coordinate":[7,6],"connectedNodes":[14,12]},'
        + '{"coordinate":[6,6],"connectedNodes":[15,13]},'
        + '{"coordinate":[5,6],"connectedNodes":[16,14,29]},'
        + '{"coordinate":[4,6],"connectedNodes":[17,15,39]},'
        + '{"coordinate":[3,6],"connectedNodes":[18,16]},'
        + '{"coordinate":[2,6],"connectedNodes":[19,30,17]},'
        + '{"coordinate":[1,6],"connectedNodes":[20,18]},'
        + '{"coordinate":[0,6],"connectedNodes":[19,21]},'
        + '{"coordinate":[0,5],"connectedNodes":[20,22]},'
        + '{"coordinate":[0,4],"connectedNodes":[21,31,23]},'
        + '{"coordinate":[0,3],"connectedNodes":[22,24]},'
        + '{"coordinate":[0,2],"connectedNodes":[23,25]},'
        + '{"coordinate":[0,1],"connectedNodes":[24,26]},'
        + '{"coordinate":[1,1],"connectedNodes":[25,27,32]},'
        + '{"coordinate":[2,1],"connectedNodes":[26,28]},'
        + '{"coordinate":[2,2],"connectedNodes":[7,27]},'
        + '{"coordinate":[5,5],"connectedNodes":[15]},'
        + '{"coordinate":[2,7],"connectedNodes":[18]},'
        + '{"coordinate":[1,4],"connectedNodes":[22]},'
        + '{"coordinate":[1,0],"connectedNodes":[26]},'
        + '{"coordinate":[9,5],"connectedNodes":[11]},'
        + '{"coordinate":[5,0],"connectedNodes":[5]},'
        + '{"coordinate":[4,2],"connectedNodes":[7,36,8]},'
        + '{"coordinate":[4,3],"connectedNodes":[37,38,35]},'
        + '{"coordinate":[3,3],"connectedNodes":[36,7]},'
        + '{"coordinate":[4,4],"connectedNodes":[39,36]},'
        + '{"coordinate":[4,5],"connectedNodes":[40,16,38]},'
        + '{"coordinate":[3,5],"connectedNodes":[41,39]},'
        + '{"coordinate":[2,5],"connectedNodes":[40,42]},'
        + '{"coordinate":[2,4],"connectedNodes":[41]}]'
    )
    level74.traffic_lights = "[]"
    level74.destinations = "[[2,4]]"
    level74.origin = '{"coordinate":[9,2],"direction":"W"}'
    level74.model_solution = "[13]"

    level74.save()

    LevelDecor.objects.filter(level=level74).delete()

    set_decor(
        level74,
        json.loads(
            '[{"x":738,"y":33,"decorName":"tree1"},'
            + '{"x":496,"y":404,"decorName":"tree2"},'
            + '{"x":248,"y":39,"decorName":"pond"},'
            + '{"x":101,"y":500,"decorName":"bush"},'
            + '{"x":148,"y":561,"decorName":"bush"},'
            + '{"x":97,"y":565,"decorName":"bush"}]'
        ),
    )
    set_blocks(
        level74,
        json.loads(
            '[{"type":"move_forwards","number":1},'
            + '{"type":"turn_left","number":2},'
            + '{"type":"turn_right","number":1},'
            + '{"type":"turn_around","number":2},'
            + '{"type":"controls_repeat","number":1},'
            + '{"type":"call_proc"},{"type":"declare_proc"}]'
        ),
    )

    level73.model_solution = "[15]"
    level75.model_solution = "[10, 26]"

    level75.name = "73"
    level73.name = "75"

    level72.save()
    level75.save()
    level74.save()
    level73.save()
    level79.save()

    level88 = Level.objects.get(name="88", default=1)
    levelBlock = LevelBlock(level=level88, type=Block.objects.get(type="wait"))
    levelBlock.save()

    level89 = Level.objects.get(name="89", default=1)
    levelBlock = LevelBlock(level=level89, type=Block.objects.get(type="logic_negate"))
    levelBlock.save()

    levelBlock = LevelBlock.objects.get(
        type=Block.objects.get(type="controls_repeat_until"), level=level89
    )
    levelBlock.delete()

    level90 = Level.objects.get(name="90", default=1)
    level90.path = (
        '[{"coordinate":[1,6],"connectedNodes":[1]},'
        + '{"coordinate":[2,6],"connectedNodes":[0,2]},'
        + '{"coordinate":[2,5],"connectedNodes":[1,3]},'
        + '{"coordinate":[3,5],"connectedNodes":[2,4]},'
        + '{"coordinate":[4,5],"connectedNodes":[3,5]},'
        + '{"coordinate":[4,4],"connectedNodes":[4,6]},'
        + '{"coordinate":[5,4],"connectedNodes":[5,7]},'
        + '{"coordinate":[5,3],"connectedNodes":[6,8]},'
        + '{"coordinate":[6,3],"connectedNodes":[7,9]},'
        + '{"coordinate":[7,3],"connectedNodes":[8,10]},'
        + '{"coordinate":[7,2],"connectedNodes":[9,11]},'
        + '{"coordinate":[8,2],"connectedNodes":[10,12]},'
        + '{"coordinate":[8,1],"connectedNodes":[13,11]},'
        + '{"coordinate":[7,1],"connectedNodes":[14,12]},'
        + '{"coordinate":[6,1],"connectedNodes":[15,13]},'
        + '{"coordinate":[5,1],"connectedNodes":[16,14]},'
        + '{"coordinate":[4,1],"connectedNodes":[17,15]},'
        + '{"coordinate":[3,1],"connectedNodes":[18,16]},'
        + '{"coordinate":[3,2],"connectedNodes":[19,17]},'
        + '{"coordinate":[2,2],"connectedNodes":[20,18]},'
        + '{"coordinate":[2,3],"connectedNodes":[21,19]},'
        + '{"coordinate":[1,3],"connectedNodes":[22,20]},'
        + '{"coordinate":[1,4],"connectedNodes":[23,21]},'
        + '{"coordinate":[0,4],"connectedNodes":[22]}]'
    )
    level90.destinations = "[[0,4]]"
    level90.save()

    level101 = Level.objects.get(name="101", default=1)
    level101.blocklyEnabled = False
    level101.save()

    level105 = Level.objects.get(name="105", default=1)
    Level.objects.filter(name="106", default=1).exclude(
        pk=level105.next_level.id
    ).delete()

    level106 = Level.objects.get(name="106", default=1)
    Level.objects.filter(name="107", default=1).exclude(
        pk=level106.next_level.id
    ).delete()


class Migration(migrations.Migration):

    dependencies = [("game", "0001_squashed_0025_levels_ordering_pt1")]

    operations = [migrations.RunPython(change_levels)]



"""
    Character data
"""

from builtins import object
from rest_framework.reverse import reverse


class Character(object):
    def __init__(self, pk, name, en_face, top_down, width, height):
        self.id = self.pk = pk
        self.name = name
        self.en_face = en_face
        self.top_down = top_down
        self.width = width
        self.height = height


CHARACTER_DATA = {
    "Van": Character(
        pk=1,
        name=u"Van",
        # FUTURE: add external branding option
        en_face=u"characters/front_view/Van.svg",
        top_down=u"characters/top_view/Van.svg",
        height="20",
        width="40",
    ),
    "Dee": Character(
        pk=2,
        name=u"Dee",
        # FUTURE: add external branding option
        en_face=u"characters/front_view/Dee.svg",
        top_down=u"characters/top_view/Dee.svg",
        height="28",
        width="52",
    ),
    "Nigel": Character(
        pk=3,
        name=u"Nigel",
        en_face=u"characters/front_view/Nigel.svg",
        top_down=u"characters/top_view/Nigel.svg",
        height="32",
        width="56",
    ),
    "Kirsty": Character(
        pk=4,
        name=u"Kirsty",
        en_face=u"characters/front_view/Kirsty.svg",
        top_down=u"characters/top_view/Kirsty.svg",
        height="32",
        width="60",
    ),
    "Wes": Character(
        pk=5,
        name=u"Wes",
        en_face=u"characters/front_view/Wes.svg",
        top_down=u"characters/top_view/Wes.svg",
        height="20",
        width="40",
    ),
    "Phil": Character(
        pk=6,
        name=u"Phil",
        en_face=u"characters/front_view/Phil.svg",
        top_down=u"characters/top_view/Phil.svg",
        height="40",
        width="40",
    ),
}


def get_character(name):
    """Helper method to get a character."""
    return CHARACTER_DATA[name]


def get_all_character():
    return list(CHARACTER_DATA.values())


def get_character_by_pk(pk):
    for character in list(CHARACTER_DATA.values()):
        if character.pk == int(pk):
            return character
    raise KeyError


def get_characters_url(pk, request):
    return reverse("character-detail", args={pk}, request=request)



from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [("game", "0028_level_disable_route_score")]

    operations = [
        migrations.AddField(
            model_name="level",
            name="pythonViewEnabled",
            field=models.BooleanField(default=False),
            preserve_default=True,
        )
    ]



from django.db import migrations


def update_level(apps, schema_editor):

    Level = apps.get_model("game", "Level")
    level50 = Level.objects.get(name="50", default=1)
    level50.model_solution = "[11]"
    level50.destinations = "[[7,4]]"
    level50.save()


def dummy_reverse(apps, schema_editor):
    """It's not possible to reverse this data migration
    but we want to allow Django to reverse previous migrations.
    """
    pass


class Migration(migrations.Migration):
    dependencies = [("game", "0071_use_common_models")]
    operations = [migrations.RunPython(update_level, reverse_code=dummy_reverse)]
